import importlib
import inspect
import logging
from importlib.util import find_spec

from anilius.core.module import Module
from anilius.core.service import Service
from anilius.core.settings import settings
from anilius.server.grpcio import GRPCIOServer
from anilius.utils.singleton import Singleton
from prometheus_client import start_http_server

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class App(metaclass=Singleton):
    _modules = {}

    def __init__(self):
        self.service = Service()

        self.enable_modules = settings.get("ENABLE_MODULES", [])

        assert (
                type(self.enable_modules) is tuple or type(self.enable_modules) is list
        ), "ENABLE_MODULES should be iterable"

        self._initialize_modules()

        self.server = GRPCIOServer([], self.service)

        start_http_server(8081)
        # sentry_init(dsn=settings.get("SENTRY_DSN", ""))

    def _initialize_module(self, module):
        self._modules[module.name] = module()

    def _initialize_modules(self):
        root_of_module = "modules"

        for module in self.enable_modules:
            module_path = root_of_module + "." + module + ".module"

            assert find_spec(module_path) is not None, "Enabled modules should exist"

            mod = importlib.import_module(module_path)

            for entity_key in dir(mod):
                entity = getattr(mod, entity_key)

                if (
                        inspect.isclass(entity)
                        and issubclass(entity, Module)
                        and entity is not Module
                ):
                    self._initialize_module(entity)

    @staticmethod
    async def run():
        app = App()
        await app.server.start()



import os
import sys

sys.path.append(os.getcwd())


def main(as_module=False):
    from anilius.cli.anilius import cli

    cli.main(args=sys.argv[1:], prog_name="python -m anilius" if as_module else None)


if __name__ == "__main__":
    main(as_module=True)



from anilius.core.serializer_field import SerializerField


class StringField(SerializerField):
    def validate(self):
        return type(self._raw_value) is str

    def get_value(self):
        return str(self._raw_value)



from collections import OrderedDict

from anilius.core.serializer_field import SerializerField
from google.protobuf.message import Message
from google.protobuf.internal.containers import RepeatedScalarFieldContainer


class SerializerMetaclass(type):
    """
    This metaclass sets a dictionary named `_declared_fields` on the class.

    Any instances of `Field` included as attributes on either the class
    or on any of its superclasses will be include in the
    `_declared_fields` dictionary.
    """

    @classmethod
    def _get_declared_fields(mcs, bases, attrs):
        fields = [
            (field_name, attrs.pop(field_name))
            for field_name, obj in list(attrs.items())
            if isinstance(obj, SerializerField)
        ]
        fields.sort(key=lambda x: x[1].get_creation_counter())

        # Ensures a base class field doesn't override cls attrs, and maintains
        # field precedence when inheriting multiple parents. e.g. if there is a
        # class C(A, B), and A and B both define 'field', use 'field' from A.
        known = set(attrs)

        def visit(name):
            known.add(name)
            return name

        base_fields = [
            (visit(name), f)
            for base in bases
            if hasattr(base, "_declared_fields")
            for name, f in getattr(base, "_declared_fields").items()
            if name not in known
        ]

        return OrderedDict(base_fields + fields)

    def __new__(mcs, name, bases, attrs):
        attrs["_declared_fields"] = mcs._get_declared_fields(bases, attrs)
        return super().__new__(mcs, name, bases, attrs)


class Serializer(SerializerField, metaclass=SerializerMetaclass):
    _declared_fields = None

    def __init__(self, request):
        super().__init__()
        assert isinstance(request, Message), "Request should be type of Message"

        for field in request.ListFields():
            if field[0].name in self._declared_fields:
                raw_value = getattr(request, field[0].name)
                raw_value = self.extract_message(raw_value)
                self._declared_fields[field[0].name].set_raw_value(raw_value)

    def extract_message(self, raw_value):
        if isinstance(raw_value, Message):
            raw_dict = {}
            for field in raw_value.ListFields():
                raw_value = getattr(raw_value, field[0].name)
                raw_value = self.extract_message(raw_value)
                raw_dict[field[0].name] = raw_value
            raw_value = raw_dict
        elif type(raw_value) is RepeatedScalarFieldContainer:
            raw_list = []

            for element in raw_value:
                raw_list.append(self.extract_message(element))

            raw_value = raw_list

        print(type(raw_value))

        return raw_value

    def validate(self):
        return True

    def get_value(self):
        return self.to_dict()

    def get_declared_fields(self):
        return self._declared_fields

    def to_dict(self):
        return dict(self.get_declared_fields())



from contextlib import AbstractContextManager

from anilius.db.db import DB
from anilius.utils.singleton import Singleton
from sqlalchemy.orm import scoped_session, sessionmaker


class DBSession(metaclass=Singleton):
    def __init__(
            self,
            autocommit=False,
            autoflush=False,
            commit=True,
            expire_on_commit=False,
            close=True,
    ):
        self.engine = DB.get_engine()
        self.session = scoped_session(
            sessionmaker(
                autocommit=autocommit,
                autoflush=autoflush,
                bind=self.engine,
                expire_on_commit=expire_on_commit,
            )
        )
        self.model = DB.get_model()
        self.model.query = self.session.query_property()
        self.commit = commit
        self.close = close

    def __call__(self, commit=True, close=True):
        self.commit = commit
        self.close = close

    def __enter__(self):
        return self.session

    def __del__(self):
        self.session.rollback()
        self.session.close()

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):

        if exc_val:
            self.session.rollback()
            self.session.close()
            return False
        elif self.commit:
            self.session.commit()
            if self.close:
                self.session.close()

        return True



from anilius.cache.compressors.lz4 import Lz4Compressor
from anilius.cache.exceptions import CompressorError
from anilius.cache.serializers.msgpack import MSGPackSerializer
from anilius.redis.client import RedisClient


class Cache:
    _compressor = Lz4Compressor()
    _serializer = MSGPackSerializer()

    def decode(self, value):
        """
        Decode the given value.
        """

        try:
            value = int(value)
        except (ValueError, TypeError):
            try:
                value = self._compressor.decompress(value)
            except CompressorError:
                # Handle little values, chosen to be not compressed
                pass
            value = self._serializer.loads(value)
        return value

    def encode(self, value):
        """
        Encode the given value.
        """

        if isinstance(value, bool) or not isinstance(value, int):
            value = self._serializer.dumps(value)
            value = self._compressor.compress(value)
            return value

        return value

    @staticmethod
    def set(*args, **kwargs):
        with RedisClient as client:
            return client.set(*args, **kwargs)

    @staticmethod
    def get(key, default=None, version=None, _client=None):
        with RedisClient as client:
            return client.get(key, default=default, version=version, _client=_client)

    @staticmethod
    def delete(*args, **kwargs):
        with RedisClient as client:
            return client.delete(*args, **kwargs)

    @staticmethod
    def incr(*args, **kwargs):
        with RedisClient as client:
            return client.incr(*args, **kwargs)

    @staticmethod
    def decr(*args, **kwargs):
        with RedisClient as client:
            return client.decr(*args, **kwargs)

    @staticmethod
    def keys(*args, **kwargs):
        with RedisClient as client:
            return client.keys(*args, **kwargs)

    @staticmethod
    def ttl(*args, **kwargs):
        with RedisClient as client:
            return client.ttl(*args, **kwargs)

    @staticmethod
    def persist(*args, **kwargs):
        with RedisClient as client:
            return client.persist(*args, **kwargs)

    @staticmethod
    def expire(*args, **kwargs):
        with RedisClient as client:
            return client.expire(*args, **kwargs)

    @staticmethod
    def lock(*args, **kwargs):
        with RedisClient as client:
            return client.lock(*args, **kwargs)

    @staticmethod
    def close(**kwargs):
        with RedisClient as client:
            client.close(**kwargs)

    @staticmethod
    def touch(*args, **kwargs):
        with RedisClient as client:
            return client.touch(*args, **kwargs)



from anilius.core.service import Service


class ServiceWrapper:
    proto_service_path = None

    def __init__(self, proto_service_path):
        self.proto_service_path = proto_service_path
        self.service = Service()

    def add(self, rpc_method, controller):
        self.service.add(self.proto_service_path, rpc_method, controller)



import importlib

import grpc
from anilius.utils.singleton import Singleton
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

_sym_db = _symbol_database.Default()


class Service(metaclass=Singleton):
    _services = {}

    @staticmethod
    def _get_services_from_proto(proto_service_path):
        proto_service_module = importlib.import_module(proto_service_path)

        assert hasattr(
            proto_service_module, "DESCRIPTOR"
        ), "Should gave valid proto service path"

        descriptor = getattr(proto_service_module, "DESCRIPTOR")

        assert hasattr(
            descriptor, "services_by_name"
        ), "Should gave valid proto service path"

        services = getattr(descriptor, "services_by_name")

        explicit_services = list()
        for service in services:
            explicit_services.append(services[service])

        return explicit_services

    def _add_rpc_handler(self, service, rpc_method, controller):
        rpc_handler = self._create_rpc_handler(rpc_method, controller)

        assert (
                service not in self._services
                or rpc_method.name not in self._services[service]
        ), "{} rpc for {} service is already add".format(service, rpc_method.name)

        if service in self._services:
            self._services[service][rpc_method.name] = rpc_handler
        else:
            self._services[service] = {rpc_method.name: rpc_handler}

    @staticmethod
    def _create_controller_wrapper(controller, request_deserializer, response_serializer):
        def wrapper(request, context):
            return controller(request, context, request_deserializer, response_serializer).get_response()

        return wrapper

    def _create_rpc_handler(self, rpc_method, controller):
        request_deserializer = self._create_reflection_from_method(
            rpc_method.input_type
        )
        response_serializer = self._create_reflection_from_method(
            rpc_method.output_type
        )

        handler = grpc.unary_unary_rpc_method_handler(
            self._create_controller_wrapper(controller, request_deserializer, response_serializer),
            request_deserializer=request_deserializer.FromString,
            response_serializer=response_serializer.SerializeToString,
        )

        return handler

    @staticmethod
    def _convert_file_to_module(file_name):
        return file_name.split(".")[0].replace("/", ".") + "_pb2"

    def _create_reflection_from_method(self, descriptor):
        reflection = _reflection.GeneratedProtocolMessageType(
            descriptor.name,
            (_message.Message,),
            {
                "DESCRIPTOR": descriptor,
                "__module__": self._convert_file_to_module(descriptor.file.name),
            },
        )
        _sym_db.RegisterMessage(reflection)

        return reflection

    def get_generic_handlers(self):
        generic_handlers = []

        for service in self._services:
            generic_handler = grpc.method_handlers_generic_handler(
                service, self._services[service]
            )
            generic_handlers.append(generic_handler)

        return generic_handlers

    def add_handlers_to_server(self, server):
        generic_handlers = self.get_generic_handlers()
        server.add_generic_rpc_handlers(generic_handlers)

    def add(self, proto_service_path, rpc_method, controller):
        services = self._get_services_from_proto(proto_service_path)

        for service in services:
            methods = getattr(service, "methods_by_name")

            if rpc_method in methods:
                self._add_rpc_handler(
                    service.full_name, methods[rpc_method], controller
                )
                break



import platform

import click
from grpc import __version__ as grpc_version
from sqlalchemy import __version__ as sqlalchemy_version


def get_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return

    from anilius import __version__

    message = "Python %(python)s\nAnilius %(anilius)s\ngRPC %(grpc)s\nSQLAlchemy %(sqlalchemy)s"
    click.echo(
        message
        % {
            "python": platform.python_version(),
            "anilius": __version__,
            "grpc": grpc_version,
            "sqlalchemy": sqlalchemy_version,
        },
        color=ctx.color,
    )
    ctx.exit()


version_option = click.Option(
    ["--version"],
    help="Show the anilius version",
    expose_value=False,
    callback=get_version,
    is_flag=True,
    is_eager=True,
)



from abc import abstractmethod

from anilius.core.settings import settings


class BaseServe(object):
    server = None
    service = None
    interceptors = None
    host = "0.0.0.0"
    port = "8080"

    def __init__(self, interceptors, service):
        self.host = settings.get("GRPC_SERVER_HOST", "0.0.0.0")
        self.port = settings.get("GRPC_SERVER_PORT", "8080")

        self.service = service
        self.interceptors = interceptors

    @property
    def _address(self):
        return self.host + ":" + self.port

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass



import jwt

from anilius.core.settings import settings


def encode_jwt(claim):
    return jwt.encode(claim, settings.get("JWT_SECRET", ""), algorithm="HS512")


def decode_jwt(encoded):
    return jwt.decode(encoded, settings.get("JWT_SECRET", ""), algorithms=["HS512"])



import signal
import threading
from concurrent import futures

import grpc
from anilius.server.server import BaseServe

NUM_SECS_TO_WAIT = 10


class GRPCIOServer(BaseServe):

    def __init__(self, interceptors, service):
        super().__init__(interceptors, service)

        self.server = self.server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=10),
            interceptors=interceptors,
        )

        self.event_shutting_down = threading.Event()

        self.service.add_handlers_to_server(self.server)
        self.server.add_insecure_port(self._address)

        signal.signal(signal.SIGTERM, self.on_stop)

    def on_stop(self, signum, frame):
        self.event_shutting_down.set()

    def graceful_stop(self):
        self.server.stop(NUM_SECS_TO_WAIT).wait()
        self.server.wait_for_termination()

    def run(self):
        try:
            self.server.start()
            self.event_shutting_down.wait()
            self.graceful_stop()
        except KeyboardInterrupt:
            self.graceful_stop()

    async def start(self):
        self.run()

    def stop(self):
        self.graceful_stop()



import uuid
from typing import Dict

from anilius.db.db import Model
from anilius.db.db_session import DBSession
from sqlalchemy import Column, DateTime
from sqlalchemy.sql import func

from anilius.model.uuid import UUID
from anilius.utils.uuid import validate_uuid4


class BaseModel(Model):
    __abstract__ = True

    id = Column(UUID(), primary_key=True, default=uuid.uuid4)
    created_on = Column(DateTime, default=func.now())
    updated_on = Column(DateTime, default=func.now(), onupdate=func.now())

    @classmethod
    def create(cls, commit=True, close=True, **entity_properties):
        entity = cls(**entity_properties)
        with DBSession(commit=commit, close=close) as session:
            session.add(entity)

        return entity

    @classmethod
    def get(cls, *entity_id):
        for enttid in entity_id:
            if not validate_uuid4(enttid):
                raise ValueError("entity id should be uuid4")
        with DBSession(commit=False) as session:
            entities = [session.query(cls).get(enttid) for enttid in entity_id]
        return tuple(entities)

    @classmethod
    def list(cls, condition_dict):
        with DBSession(commit=False) as session:
            entity = session.query(cls).filter(*condition_dict).all()
        return entity

    @classmethod
    def delete(cls, **entity_properties) -> None:
        """deletes the given entry from database table"""

        with DBSession(commit=False) as session:
            entity = session.query(cls).filter_by(**entity_properties).first()
            session.delete(entity)



import msgpack

from anilius.cache.serializers.base import BaseSerializer


class MSGPackSerializer(BaseSerializer):
    def dumps(self, value):
        return msgpack.dumps(value)

    def loads(self, value):
        return msgpack.loads(value, raw=False)



from anilius.core.serializer_field import SerializerField


class IntegerField(SerializerField):
    def validate(self):
        return type(self._raw_value) is int

    def get_value(self):
        if type(self._raw_value) is int:
            return self._raw_value
        elif self._raw_value is None:
            return 0

        return int(self._raw_value)



from abc import ABC, abstractmethod

from anilius.db.db import DB
from anilius.wrappers.service import ServiceWrapper


class Module(ABC):
    service_proto_module = None
    name = None
    service_wrapper = None
    db_wrapper = None

    def __init__(self):
        if self.name is None:
            self.name = self.__class__.__name__

        assert (
            self.service_proto_module is not None
        ), "Should gave valid proto path of service descriptor"

        self.service_wrapper = ServiceWrapper(self.service_proto_module)
        self.db = DB

        self.initialise(self.service_wrapper, self.db)

    @abstractmethod
    def initialise(self, service, db):
        raise NotImplemented



import threading


class Singleton(type):
    _instances = {}
    _locks = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            if cls.__name__ not in cls._locks:
                cls._locks[cls.__name__] = threading.Lock()

            with cls._locks[cls.__name__]:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(
                        *args, **kwargs
                    )
        return cls._instances[cls]



class SerializerField:
    _creation_counter = 0
    _raw_value = None

    def __init__(self):
        self._creation_counter = SerializerField._creation_counter
        SerializerField._creation_counter += 1

    def get_creation_counter(self):
        return self._creation_counter

    def validate(self):
        raise NotImplementedError()

    def get_value(self):
        raise NotImplementedError()

    def set_raw_value(self, raw):
        self._raw_value = raw



import re

UUID4_PATTERN = (
    "^[0-9a-f]{8}-[0-9a-f]{4}-[1-5][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}$"
)


def validate_uuid4(uid):
    uid = str(uid)
    return bool(re.match(UUID4_PATTERN, uid))



import string
import random

RANDOM_STRING_CHOICES = string.ascii_lowercase + string.digits
RANDOM_STRING_DIGITS_CHOICES = string.digits


def random_string(specific_len):
    return "".join(random.choice(RANDOM_STRING_CHOICES) for _ in range(specific_len))


def random_string_digits(specific_len):
    return "".join(
        random.choice(RANDOM_STRING_DIGITS_CHOICES) for _ in range(specific_len)
    )



import json

from anilius.cache.serializers.base import BaseSerializer


class JSONSerializer(BaseSerializer):
    def dumps(self, value):
        return json.dumps(value).encode()

    def loads(self, value):
        return json.loads(value.decode())



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base

from anilius.core.settings import settings
from anilius.utils.singleton import Singleton


class DB(metaclass=Singleton):
    engine = None
    Model = None

    def __init__(self):
        self.engine = create_engine(
            settings.get("DATABASE_URI", "mysql://127.0.0.1:3306/anilius"),
            **settings.get("DATABASE_CONNECT_OPTIONS", {})
        )
        self.Model = declarative_base(name="Model")

    def get_model(self):
        return self.Model

    def get_engine(self):
        return self.engine

    def add(self, model):
        pass


DB = DB()
Model = DB.get_model()



from abc import ABC

import grpc
from anilius.core.permission import Permission
from anilius.core.serializer import Serializer
from anilius.core.serializer_field import SerializerField
from anilius.utils.jwt import decode_jwt
from jwt import InvalidAlgorithmError, InvalidSignatureError


class Controller(ABC):
    permissions = ()
    payload = {}
    is_authorize = False
    request_serializer = None
    request_deserializer = None
    response_serializer = None
    response_deserializer = None
    _serialized_data = None
    meta = {}

    def __init__(self, request, context, request_deserializer, response_serializer):
        self.request_deserializer = request_deserializer
        self.response_serializer = response_serializer
        self.request = request
        self.context = context

        self.metadata = context.invocation_metadata()

        self.parse()

    def check_permissions(self):
        for permission in self.permissions:
            has_perm = permission.has_perm(self)
            if not has_perm:
                self.raise_permission()
                break

    def raise_permission(self):
        self.context.set_code(grpc.StatusCode.PERMISSION_DENIED)
        self.context.set_details("You have not permission for this action")

    def parse(self):
        self.meta = {}
        self._serialized_data = self.get_request_serializer(self.request).to_dict()

        for data in self.metadata:
            self.meta[data.key] = data.value

        for permission in self.permissions:
            assert isinstance(
                permission, Permission
            ), "permissions should be type of Permission"

        authorization = self.meta.get("authorization", None)
        if authorization is not None:
            self.extract_payload(authorization)
        else:
            self.payload = {}
            self.is_authorize = False

    def extract_payload(self, authorization):
        try:
            self.payload = decode_jwt(authorization)
            self.is_authorize = True
        except (ValueError, InvalidAlgorithmError, InvalidSignatureError):
            pass

    def get_valid_data(self, key, default=None):
        field = self._serialized_data.get(key, None)

        if not isinstance(field, SerializerField):
            return default

        return field.get_value()

    @property
    def get_request_serializer(self):
        return (
            self.request_serializer
            if self.request_serializer is not None
            else Serializer
        )

    def get_response(self):
        self.check_permissions()



import io
import re
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with io.open("src/anilius/__init__.py", "rt", encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)

setuptools.setup(
    name="Anilius",
    version=version,
    description="A high performance grpc micro framework based on python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Malivix/Anilius",
    author="Mahdi Zarrintareh",
    author_email="mza2rintareh@gmail.com",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    include_package_data=True,
    python_requires=">=3.0",
    entry_points={"console_scripts": ["anilius = anilius.cli.main:main"]},
)



from anilius.core.serializer_field import SerializerField


class ListField(SerializerField):
    def __init__(self, list_type=None):
        super().__init__()
        self.list_type = list_type

    def validate(self):
        result = type(self._raw_value) is list

        if self.list_type and result:
            for item in self._raw_value:
                result = result and item.validate()

        return result

    def get_value(self):
        return self._raw_value



import os

import click
from anilius.cli.create import create_command
from anilius.cli.serve import serve_command
from anilius.cli.version import version_option


class AniliusGroup(click.Group):
    def __init__(
            self,
            create_app=None,
            add_version_option=True,
            load_dotenv=True,
            set_debug_flag=True,
            **extra
    ):
        params = list(extra.pop("params", None) or ())

        if add_version_option:
            params.append(version_option)

        click.Group.__init__(self, params=params, **extra)

        self.add_command(create_command)
        self.add_command(serve_command)

        self.create_app = create_app
        self.load_dotenv = load_dotenv
        self.set_debug_flag = set_debug_flag

        self._loaded_plugin_commands = False

    def main(self, *args, **kwargs):
        os.environ["ANILIUS_RUN_FROM_CLI"] = "true"

        kwargs.setdefault("auto_envvar_prefix", "ANILIUS")
        return super(AniliusGroup, self).main(*args, **kwargs)


cli = AniliusGroup(
    help="""\
A watchlist utility script for Anilius microservice.
""".format(
        cmd="export" if os.name == "posix" else "set",
        prefix="$ " if os.name == "posix" else "> ",
    )
)



"""
Settings and configuration for Django.

Read values from the module specified by the DJANGO_SETTINGS_MODULE environment
variable, and then from django.conf.global_settings; see the global_settings.py
for a list of all possible variables.
"""

import importlib
from importlib.util import find_spec


class Settings(object):
    """
    A lazy proxy for either global Django settings or a custom settings object.
    The user can manually configure settings prior to using them. Otherwise,
    Django uses the settings module pointed to by DJANGO_SETTINGS_MODULE.
    """

    def __init__(self):
        """
        Load the settings module pointed to by the environment variable. This
        is used the first time settings are needed, if the user hasn't
        configured settings manually.
        """
        settings_module = "settings"

        if find_spec(settings_module):
            mod = importlib.import_module(settings_module)

            self._explicit_settings = set()
            for setting in dir(mod):
                if setting.isupper():
                    setting_value = getattr(mod, setting)
                    setattr(self, setting, setting_value)
                    self._explicit_settings.add(setting)

    def get(self, key, default=None):
        if hasattr(self, key):
            return getattr(self, key)

        return default

    def set(self, key, value):
        self.__setattr__(key, value)

    def __setattr__(self, name, value):
        """
        Set the value of setting. Clear all cached values if _wrapped changes
        (@override_settings does this) or clear single values when set.
        """
        super().__setattr__(name, value)

    def __delattr__(self, name):
        """Delete a setting and clear it from cache if needed."""
        super().__delattr__(name)


settings = Settings()



import asyncio

import click
from anilius.core.app import App


def serve():
    asyncio.run(App.run())


serve_command = click.Command(
    "serve",
    help="Start anilius and serve",
    callback=serve,
)



import uuid

from sqlalchemy.types import TypeDecorator, BINARY
from sqlalchemy.dialects.postgresql import UUID as psqlUUID


class UUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses Postgresql's UUID type, otherwise uses
    BINARY(16), to store UUID.

    """
    impl = BINARY
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(psqlUUID())
        else:
            return dialect.type_descriptor(BINARY(16))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                if isinstance(value, bytes):
                    value = uuid.UUID(bytes=value)
                elif isinstance(value, int):
                    value = uuid.UUID(int=value)
                elif isinstance(value, str):
                    value = uuid.UUID(value)
        if dialect.name == "postgresql":
            return str(value)
        else:
            return value.bytes

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        if dialect.name == "postgresql":
            return uuid.UUID(value)
        else:
            return uuid.UUID(bytes=value)
